#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 14:41
# @Author  : 兵
# @email    : 1747193328@qq.com

from .gpumd import run_gpumd
from .nep import run_nep
from .perturb import run_perturb
from .template import init_template
from .train import train_nep
from .vasp import run_vasp
from .select import run_select



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 14:33
# @Author  : 兵
# @email    : 1747193328@qq.com
import argparse
import sys
sys.path.append('../../')
from NepTrain.core import *
from NepTrain import __version__
import warnings

# 禁用所有 UserWarning
warnings.simplefilter('ignore', UserWarning)
def check_kpoints_number(value):
    """检查值是否为单个数字或三个数字的字符串"""

    if isinstance(value, str):
        values = value.split(',')

        if len(values) == 3 and all(v.isdigit() for v in values):
            return list(map(int, values))
        elif len(values) == 1 and value.isdigit():
            return [int(value),int(value),int(value)]
        else:
            raise argparse.ArgumentTypeError("The ka parameter must be a single number or three numbers separated by `,`.")
    elif isinstance(value, int):
        return value
    else:
        raise argparse.ArgumentTypeError("The ka parameter must be a single number or three numbers separated by `,`.")

def build_init(subparsers):
    parser_init = subparsers.add_parser(
        "init",
        help="Initialize some file templates",
    )


    parser_init.add_argument("-f", "--force", action='store_true',
                             default=False,
                             help="Force overwriting of generated templates"
                             )

    parser_init.set_defaults(func=init_template)


def build_perturb(subparsers):
    parser_perturb = subparsers.add_parser(
        "perturb",
        help="Generate perturbed structures.",
    )

    parser_perturb.set_defaults(func=run_perturb)

    parser_perturb.add_argument("model_path",
                             type=str,

                             help="The structure path or structure file required for calculation only supports files in xyz and vasp formats.")
    parser_perturb.add_argument("--num","-n",
                             type=int,
                                default=20,
                             help="The number of perturbations for each structure, if a folder is input, the final number generated should be the number of structures multiplied by num.default 20.")

    parser_perturb.add_argument("--cell", "-c",
                                dest="cell_pert_fraction",
                                type=float,
                                default=0.03,
                                help="The deformation ratio,default 0.03.")

    parser_perturb.add_argument("--distance", "-d",
                                type=float,
                                dest="min_distance",
                                default=0.1,
                                help="Min atom distance, unit Å, default 0.1.")

    parser_perturb.add_argument("--out", "-o",
                             dest="out_file_path",
                             type=str,
                             help="Output file for perturbed structures, default ./perturb.xyz.",
                             default="./perturb.xyz"
                             )
    parser_perturb.add_argument("--append", "-a",
                             dest="append", action='store_true', default=False,
                             help="Write to out_file_path in append mode, default False.",

                             )
    parser_perturb.add_argument("--filter", "-f", action="store_true",
                               help="Whether to filter based on minimum bond length, default is False.",
                               default=False)
def build_vasp(subparsers):
    parser_vasp = subparsers.add_parser(
        "vasp",
        help="Calculate single-point energy using VASP.",
    )
    parser_vasp.set_defaults(func=run_vasp)

    parser_vasp.add_argument("model_path",
                             type=str,

                             help="The required structure path or structure file only supports files in xyz and vasp formats.")
    parser_vasp.add_argument("--directory", "-dir",

                             type=str,
                             help="Set the VASP calculation path. default ./cache/vasp.",
                             default="./cache/vasp"
                             )

    parser_vasp.add_argument("--out", "-o",
                             dest="out_file_path",
                             type=str,
                             help="Structure output file after calculation. default ./vasp_scf.xyz",
                             default="./vasp_scf.xyz"
                             )

    parser_vasp.add_argument("--append", "-a",
                             dest="append", action='store_true', default=False,
                             help="Write to out_file_path in append mode, default False.",

                             )
    parser_vasp.add_argument("--gamma", "-g",
                             dest="use_gamma", action='store_true', default=False,
                             help="Default to using Monkhorst-Pack k-points, add -g to use Gamma-centered k-point scheme.",

                             )
    parser_vasp.add_argument("-n", "-np",
                             dest="n_cpu",
                             default=1,
                             type=int,
                             help="Set the number of CPU cores, default 1.")

    parser_vasp.add_argument("--incar",

                             help="Input path for INCAR file, default is ./INCAR.",default="./INCAR")



    k_group = parser_vasp.add_mutually_exclusive_group(required=False)
    k_group.add_argument("--kspacing", "-kspacing",

                         type=float,
                         help="Set kspacing, which can also be defined in the INCAR template.")
    k_group.add_argument("--ka", "-ka",
                         default=[1, 1, 1],
                         type=check_kpoints_number,
                         help="ka takes 1 or 3 numbers (comma-separated), sets k-points to (k[0]/a, k[1]/b, k[2]/c). default 1.")


def build_nep(subparsers):
    parser_nep = subparsers.add_parser(
        "nep",
        help="Train potential functions using NEP.",
    )
    parser_nep.set_defaults(func=run_nep)


    parser_nep.add_argument("--directory", "-dir",

                             type=str,
                             help="Set the path for NEP calculations. default ./cache/nep",
                             default="./cache/nep"
                             )
    parser_nep.add_argument("--in", "-in",
                            dest="nep_in_path",
                             type=str,
                             help="Set the path for the nep.in file; if not present, generate it based on train.xyz. default ./nep.in",
                             default="./nep.in"
                             )

    parser_nep.add_argument("--train", "-train",
                             dest="train_path",

                             type=str,
                             help="Set the path for the train.xyz file, default  ./train.xyz.",
                             default="./train.xyz"
                             )
    parser_nep.add_argument("--test", "-test",
                             dest="test_path",
                             type=str,
                             help="Set the path for the test.xyz file, default is ./test.xyz.",
                             default="./test.xyz"
                             )
    parser_nep.add_argument("--nep", "-nep",
                            dest="nep_txt_path",
                             type=str,
                             help="restart and prediction require the use of a potential function, default is ./nep.txt.",
                             default="./nep.txt"
                             )
    parser_nep.add_argument("--prediction", "-pred","--pred",

                             action="store_true",
                             help="Set the forecast mode，default False",
                             default=False
                             )
    parser_nep.add_argument("--restart_file", "-restart","--restart",

                            type=str,

                            help="To restart running, simply provide a valid path; default is None.",
                             default=None
                             )
    parser_nep.add_argument("--continue_step", "-cs",

                            type=int,

                            help="If a restart_file is provided, this parameter will take effect, continuing for continue_step steps, with a default value of 10000.",
                             default=10000
                             )
def build_gpumd(subparsers):
    parser_gpumd = subparsers.add_parser(
        "gpumd",
        help="run molecular dynamics using GPUMD.",
    )
    parser_gpumd.set_defaults(func=run_gpumd)

    parser_gpumd.add_argument("model_path",
                             type=str,

                             help="The required structure path or structure file only supports files in xyz and vasp formats.")
    parser_gpumd.add_argument("--directory", "-dir",

                             type=str,
                             help="Set the GPUMD calculation path, default is ./cache/gpumd.",
                             default="./cache/gpumd"
                             )
    parser_gpumd.add_argument("--in","-in",dest="run_in_path", type=str,
                              help="The filename for the command template file, default is ./run.in.", default="./run.in")

    parser_gpumd.add_argument("--nep", "-nep",
                            dest="nep_txt_path",
                             type=str,
                             help="Potential function path, default is ./nep.txt.",
                             default="./nep.txt"
                             )
    parser_gpumd.add_argument("--time", "-t", type=int, help="Molecular dynamics time, unit ps, default 10 ps.", default=10)
    parser_gpumd.add_argument("--temperature", "-T", type=int, help="Molecular dynamics temperature in Kelvin,multiple integers can be input. default is 300 K", nargs="*", default=[300])
    parser_gpumd.add_argument("--filter", "-f", action="store_true",
                               help="Whether to filter based on minimum bond length, default is False.",
                               default=False)
    parser_gpumd.add_argument("--out", "-o",
                               dest="out_file_path",

                               type=str,
                               default="./trajectory.xyz",
                               help="Output path for structures."
                               )

def build_train(subparsers):
    parser_train = subparsers.add_parser(
        "train",
        help="Automatic training.",
    )
    parser_train.set_defaults(func=train_nep)

    parser_train.add_argument("config_path",
                             type=str,

                             help="The required structure path or structure file only supports files in XYZ and VASP formats.")


def build_select(subparsers):
    parser_select = subparsers.add_parser(
        "select",
        help="Select samples.",
    )
    parser_select.set_defaults(func=run_select,decomposition='pca')

    parser_select.add_argument("trajectory_paths",
                              nargs="+",
                             help="The trajectory files needed for sampling is in xyz format.")


    parser_select.add_argument("--base", "-base",
                               type=str,
                               default="base",
                               help="Provide a path to base.xyz, and sample the trajectory based on base.xyz."
                               )
    parser_select.add_argument("--nep", "-nep",
                               type=str,
                               default="./nep.txt",
                               help="Provide a path to a nep.txt file to extract descriptors for the structure, default is ./nep.txt. If the file does not exist, use SOAP descriptors."
                               )
    parser_select.add_argument("--max_selected", "-max", type=int,
                               help="Maximum number of structures to select, default is 20.",
                               default=20)
    parser_select.add_argument("--min_distance","-d", type=float,
                               help="Minimum bond length for farthest-point sampling, default is 0.01.",
                               default=0.01)


    dc_group = parser_select.add_mutually_exclusive_group(required=False)
    dc_group.add_argument('-pca',"--pca", action='store_const', const='pca', dest='decomposition',
                       help='Use PCA for decomposition')
    dc_group.add_argument('-umap',"--umap", action='store_const', const='umap', dest='decomposition',
                       help='Use UMAP for decomposition')

    parser_select.add_argument("--out", "-o",
                               dest="out_file_path",

                               type=str,
                               default="./selected.xyz",
                               help="Output path for selected structures.default ./selected.xyz"
                               )

    group= parser_select.add_argument_group("SOAP","SOAP Parameters")

    group.add_argument("--r_cut", "-r", type=float, help="A cutoff for local region in angstroms,default 6", default=6)
    group.add_argument("--n_max", "-n", type=int, help="The number of radial basis functions,default 8", default=8)
    group.add_argument("--l_max", "-l", type=int, help="The maximum degree of spherical harmonics,default 6", default=6)

def main():
    parser = argparse.ArgumentParser(
        description="""
        NepTrain is a tool for automatically training NEP potential functions""",

    )
    parser.add_argument(
        "-v", "--version", action="version", version=__version__
    )



    subparsers = parser.add_subparsers()


    build_init(subparsers)

    build_perturb(subparsers)

    build_select(subparsers)

    build_vasp(subparsers)

    build_nep(subparsers)
    build_gpumd(subparsers)
    build_train(subparsers)



    try:
        import argcomplete

        argcomplete.autocomplete(parser)
    except ImportError:

        pass


    args = parser.parse_args()

    try:
        _ = args.func
    except AttributeError as exc:
        parser.print_help()
        raise SystemExit("Please specify a command.") from exc
    return args.func(args)


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 15:42
# @Author  : 兵
# @email    : 1747193328@qq.com

import os
import re
import shutil
import subprocess

from rich.progress import Progress

from watchdog.events import FileSystemEventHandler


from NepTrain import utils, Config, observer
from .utils import read_symbols_from_file




class NepFileMoniter(FileSystemEventHandler):
    def __init__(self,file_path,total):

        self.file_path = file_path
        self.progress = Progress( )
        self.current_steps=0
        self.total=int(total)
        self.progress.start()
        self.pbar=self.progress.add_task(total=int(total),description="NEP training")
    def on_modified(self, event):

        if not utils.is_diff_path(event.src_path , self.file_path):
            with open(self.file_path,'r',encoding="utf8") as f:
                lines = f.readlines()
                if not lines:
                    return
                last_line=lines[-1]
                current_steps=int(last_line.split(" ")[0])

                self.progress.advance(self.pbar,current_steps-self.current_steps)
                self.current_steps=current_steps

    def finish(self):

        if self.progress.finished:
            self.progress.advance(self.pbar,self.total-self.current_steps)


        self.progress.stop()



class RunInput:

    def __init__(self,train_xyz_path,nep_in_path=None,test_xyz_path=None):
        self.nep_in_path = nep_in_path
        self.train_xyz_path = train_xyz_path
        self.test_xyz_path = test_xyz_path
        self.run_in={"generation":100000}


        self.restart=False
        if self.nep_in_path is not None and os.path.exists(self.nep_in_path):
            self.read_run(self.nep_in_path)
        self.command=Config.get('environ','nep_path')

    def read_run(self,file_name):
        with open(file_name,'r',encoding="utf8") as f:
            # groups=re.findall("(\w+)\s+(.*?)\n",f.read()+"\n")
            groups=re.findall("^([A-Za-z_]+)\s+(.*)",f.read() ,re.MULTILINE)

            for group in groups:
                self.run_in[group[0].strip()]=group[1].strip()

    def set_restart(self,file_path,steps):
        if file_path and os.path.exists(file_path):
            self.restart_nep_path=file_path
            self.run_in["generation"]=steps
            self.run_in["lambda_1"]=0
            self.restart=True


    def build_run(self):
        """
        如果runin 不存在 就遍历训练集  然后找出所有的元素

        :return:
        """
        symbols = read_symbols_from_file(self.train_xyz_path)
        self.run_in["type"]=f"{len(symbols)} {' '.join(symbols)}"

    def write_run(self,file_name):
        if  "type" not in   self.run_in :
            self.build_run()
        with open(file_name,'w',encoding="utf8") as f:
            for k,v in self.run_in.items():

                f.write(f"{k}     {v}\n" )


    def calculate(self,directory,show_progress=True):
        utils.verify_path(directory)
        if self.restart:
            utils.print_tip("Start the restart mode!")
            shutil.copy(self.restart_nep_path,os.path.join(directory,"nep.restart"))


        self.write_run(os.path.join(directory,"nep.in"))
        if self.train_xyz_path is   None or not  os.path.exists(self.train_xyz_path):
            raise ValueError("A valid train.xyz must be specified.")
        if utils.is_diff_path(self.train_xyz_path ,os.path.join(directory,"train.xyz")):

            shutil.copy(self.train_xyz_path,os.path.join(directory,"train.xyz"))
        if self.test_xyz_path is not None and os.path.exists(self.test_xyz_path):
            shutil.copy(self.test_xyz_path, os.path.join(directory, "test.xyz"))
        if show_progress:

            handler=NepFileMoniter(os.path.join(directory,"loss.out"),self.run_in["generation"])
            watch=observer.schedule(handler, os.path.abspath(directory) , recursive=False)


            if not observer.is_alive():

                observer.start()

        with   open(os.path.join(directory,"nep.out"), "w") as f_std, open(os.path.join(directory,"nep.err"), "w", buffering=1) as f_err:

            errorcode = subprocess.call(self.command,
                                        shell=True,
                                        stdout=f_std,
                                        stderr=f_err,
                                        cwd=directory)


        if show_progress:

            handler.finish()
            observer.unschedule(watch)
            observer.stop()

class PredictionRunInput(RunInput):
    def __init__(self,nep_txt_path,*args,**kwargs):
        self.nep_txt_path=nep_txt_path
        super().__init__(*args,**kwargs)

    def write_run(self,file_name):

        self.run_in["prediction"]=1
        super().write_run(file_name)

    def calculate(self,directory,show_progress=False ):
        utils.verify_path(directory)

        if self.nep_txt_path is not None and os.path.exists(self.nep_txt_path):
            if utils.is_diff_path(self.nep_txt_path, os.path.join(directory, "nep.txt")):

                shutil.copy(self.nep_txt_path, os.path.join(directory, "nep.txt"))
        else:
            raise ValueError("In prediction mode, a potential function must be specified, please specify it via the `--nep nep_path` option.")
        super().calculate(directory,show_progress)
if __name__ == '__main__':
    run=RunInput("./train1.xyz")
    # run.read_run("./nep.in")
    run.write_run("./nep.out")
    run.calculate("./")


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/14 14:34
# @Author  : 兵
# @email    : 1747193328@qq.com

from pathlib import Path
from ase.data import chemical_symbols, atomic_numbers
import re
import os



def read_symbols_from_file(file_name):
    if os.path.exists(file_name):
        with open(file_name, 'r', encoding="utf8") as f:
            trainxyz = f.read()
        groups = re.findall("^([A-Z][a-z]?)\s+", trainxyz, re.MULTILINE)
        groups = set(groups)
        symbols = []
        for symbol in groups:
            if symbol in chemical_symbols:
                symbols.append(symbol)

        symbols = sorted(symbols, key=lambda x: atomic_numbers[x])
    else:
        symbols = []
    return symbols








#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 15:42
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import re
import shutil
import subprocess

from ase.io import write as ase_write

from NepTrain import utils, Config
from .plot import *





class RunInput:


    def __init__(self,nep_txt_path):
        self.nep_txt_path=nep_txt_path
        self.command=Config.get('environ','gpumd_path')
        self.time_step=1
        self.dump_thermo=False
        self.dump_exyz=False
        self.total_time=0
        self.run_in=[]

    def set_time_temp(self,times=None,temperature=None):
        """
        None表示不修改
        :param times: 单位ps
        :param temperature:
        :return:
        """
        if times   is not None:
            self.total_time=0
        for run_index in range(len(self.run_in)):
            run=self.run_in[run_index]
            if run[0]=="ensemble":
                if temperature is not None:

                    run[1][1]=str(temperature)
                    run[1][2]=str(temperature)
            elif run[0]=="run":
                if times is not None:

                    run[1][0]=str(int(int(times)*1000*1/self.time_step))
                    self.total_time+=int(int(times)*1000*1/self.time_step)

    def read_run(self,file_name):
        self.run_in.clear()
        with open(file_name,'r',encoding="utf8") as f:
            groups=re.findall("^([A-Za-z_]+)\s+(.*)",f.read() ,re.MULTILINE)

            for group in groups:

                key,value=group[0].strip(),group[1].strip()
                if not key:
                    continue
                if key=="time_step":
                    self.time_step=float(value)
                elif key=="dump_thermo":
                    self.dump_thermo=True
                elif key=="dump_exyz":
                    self.dump_exyz=True
                elif key=="run":
                    self.total_time+=int(value)
                self.run_in.append([key,[i for i in value.split(" ") if i.strip()]])


    def write_run(self,file_name):
        with open(file_name,'w',encoding="utf8") as f:
            for i in self.run_in:

                f.write(f"{i[0]}    {' '.join(i[1])}\n")

    def calculate(self,atoms,directory,show_progress=True):
        """
        计算所需要的文件
            nep.txt
            run.in
            model.xyz
        :param directory:
        :param show_progress:
        :return:
        """
        if not os.path.exists(directory):
            os.makedirs(directory )

        self.write_run(os.path.join(directory,"run.in"))
        ase_write(os.path.join(directory,"model.xyz"),atoms,format="extxyz")
        if self.nep_txt_path is not None and os.path.exists(self.nep_txt_path):
            if utils.is_diff_path(self.nep_txt_path, os.path.join(directory, "nep.txt")):

                shutil.copy(self.nep_txt_path, os.path.join(directory, "nep.txt"))
        else:
            raise ValueError(f"{self.nep_txt_path} is an invalid path！")

        with   open(os.path.join(directory,"gpumd.out"), "w") as f_std, open(os.path.join(directory,"gpumd.err"), "w", buffering=1) as f_err:

            errorcode = subprocess.call(self.command,
                                        shell=True,
                                        stdout=f_std,
                                        stderr=f_err,
                                        cwd=directory)

        plot_energy(os.path.join(directory,"thermo.out"),len(atoms))

if __name__ == '__main__':
    # read_thermo("1.out",80)
    run=RunInput()
    run.read_run("./run.in")
    run.set_time_temp(1,300)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/28 15:01
# @Author  : 兵
# @email    : 1747193328@qq.com


from NepTrain import utils
from .io import RunInput, PredictionRunInput
from .plot import plot_nep_result
from ..utils import check_env

def run_nep(argparse):
    check_env()
    if argparse.prediction:
        run = PredictionRunInput(argparse.nep_txt_path,argparse.train_path,argparse.nep_in_path,argparse.test_path)
    else:

        run = RunInput(argparse.train_path,argparse.nep_in_path,argparse.test_path)


    run.set_restart(argparse.restart_file,argparse.continue_step)
    run.calculate(argparse.directory)
    plot_nep_result(argparse.directory)
    utils.print_success("NEP training task completed!" )


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 18:12
# @Author  : 兵
# @email    : 1747193328@qq.com
import os.path

from ase.io import read as ase_read
from ruamel.yaml import YAML

from NepTrain import module_path, utils
from .utils import check_env


def create_vasp(force):
    if   os.path.exists("./sub_vasp.sh") and not force:
        return
    utils.print_warning("Please check the queue information and environment settings in sub_vasp.sh!")

    sub_vasp="""#! /bin/bash
#SBATCH --job-name=NepTrain
#SBATCH --nodes=1
#SBATCH --partition=cpu
#SBATCH --ntasks-per-node=64
#You can place some environment loading commands here.



#eg conda activate NepTrain

$@ 

#NepTrain vasp demo.xyz -np 64 --directory ./cache -g --incar=./INCAR --kpoints 35 -o ./result/result.xyz 
"""

    with open("./sub_vasp.sh", "w",encoding="utf8") as f:
        f.write(sub_vasp)


def create_nep(force):
    if os.path.exists("./sub_gpu.sh") and not force:
        return
    utils.print_warning("Please check the queue information and environment settings in sub_gpu.sh!")

    sub_vasp = """#! /bin/bash
#SBATCH --job-name=NepTrain-gpu
#SBATCH --nodes=1
#SBATCH --ntasks-per-node=1
#SBATCH --partition=gpu-a800
#SBATCH --gres=gpu:1
#You can place some environment loading commands here.

 
$@ """
    with open("./sub_gpu.sh", "w", encoding="utf8") as f:
        f.write(sub_vasp)



def init_template(argparse):
    if not argparse.force:
        utils.print_tip("For existing files, we choose to skip; if you need to forcibly generate and overwrite, please use -f or --force.")

    if not os.path.exists("./structure"):
        os.mkdir("./structure")
        utils.print_tip("Create the directory ./structure, please place the expanded structures that need to run MD into this folder!" )
    check_env()
    create_vasp(argparse.force)
    create_nep(argparse.force)
    if not os.path.exists("./job.yaml") or argparse.force:
        utils.print_tip("You need to check and modify the vasp_job and vasp.cpu_core in the job.yaml file.")
        utils.print_warning("You also need to check and modify the settings for GPUMD active learning in job.yaml!")

        with open(os.path.join(module_path,"core/train/job.yaml"),"r",encoding="utf8") as f:

            config = YAML().load(f  )

        if os.path.exists("train.xyz"):
            #检查下第一个结构有没有计算
            atoms=ase_read("./train.xyz",0,format="extxyz")

            if not (atoms.calc and "energy"   in atoms.calc.results):
                config["current_job"]="vasp"
                utils.print_warning("Check that the first structure in train.xyz has not been calculated; set the initial task to vasp!")
        else:
            utils.print_warning("Detected that there is no train.xyz in the current directory; please check the directory structure!")
            utils.print_tip("If there is a training set but the filename is not train.xyz, please unify the job.yaml.")


        with open("./job.yaml","w",encoding="utf8") as f:
            YAML().dump(config,f  )
    else:

        #已经存在 如果执行init  更新下
        with open(os.path.join(module_path, "core/train/job.yaml"), "r", encoding="utf8") as f:

            base_config = YAML().load(f)
        with open("./job.yaml","r",encoding="utf8") as f:
            user_config = YAML().load(f)
        job=utils.merge_yaml(base_config,user_config)
        job["version"]=base_config["version"]

        with open("./job.yaml","w",encoding="utf8") as f:
            YAML().dump(job,f  )


    if not os.path.exists("./run.in")  or argparse.force:
        utils.print_tip("Create run.in; you can modify the ensemble settings! Temperature and time will be modified by the program!")

        utils.copy(os.path.join(module_path,"core/gpumd/run.in"),"./run.in")

    utils.print_success("Initialization is complete. After checking the files, you can run `NepTrain train job.yaml` to proceed.")



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 13:37
# @Author  : 兵
# @email    : 1747193328@qq.com
"""
自动训练的逻辑
"""
import os.path

from ase.io import read as ase_read
from ase.io import write as ase_write
from ruamel.yaml import YAML

from NepTrain import utils
from .worker import LocalWorker, SlurmWorker
from ..utils import check_env


class Manager:
    def __init__(self, options):
        self.options = options
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.options):
            self.index = 0
        value = self.options[self.index]
        self.index += 1
        return value

    def set_next(self, option):
        index=self.options.index(option)
        # 设置当前索引，注意索引从0开始
        if 0 <= index < len(self.options):
            self.index = index
        else:
            raise IndexError("Index out of range.")


class PathManager:



    def __init__(self, root):
        self.root = root

    def __getattr__(self, item):
        return os.path.join(self.root, item)




class NepTrainWorker:
    pass
    def __init__(self):
        self.config={}
        self.job_list=["nep","gpumd","select","vasp","pred", ]
        self.manager=Manager(self.job_list)

    def get_worker(self):
        queue = self.config.get("queue", "local")
        if queue == "local":
            return LocalWorker()
        else:
            return SlurmWorker(os.path.abspath("./sub_vasp.sh"),os.path.abspath("./sub_gpu.sh"))

    def __getattr__(self, item):

        if item.startswith("last_"):
            item=item.replace("last_","")
            generation_path=os.path.join(os.path.abspath(self.config.get("work_path")), f"Generation-{self.generation-1}")
        else:
            generation_path=os.path.join(os.path.abspath(self.config.get("work_path")), f"Generation-{self.generation}")

        if item=="generation_path":

            return generation_path

        items= item.split("_")
        if items[0] in self.job_list:
            job_path=os.path.join(generation_path, items.pop(0))
        else:
            job_path=generation_path
        fin_path=os.path.join(job_path, "_".join(items[:-1]) )
        if items[-1]=="path":
            pass
            utils.verify_path(fin_path)
        else:
            last_underscore_index = fin_path.rfind('_')
            if last_underscore_index != -1:
                # 替换最后一个下划线为点
                fin_path = fin_path[:last_underscore_index] + '.' + fin_path[last_underscore_index + 1:]
            else:
                fin_path = fin_path

            utils.verify_path(os.path.dirname(fin_path))


        return fin_path



    @property
    def generation(self):
        return self.config.get("generation")
    @generation.setter
    def generation(self,value):
        self.config["generation"] = value



    def split_vasp_job_xyz(self,xyz_file):
        addxyz = ase_read(xyz_file, ":", format="extxyz")

        split_addxyz_list = utils.split_list(addxyz, self.config["vasp_job"])

        for i, xyz in enumerate(split_addxyz_list):
            if xyz:
                ase_write(self.__getattr__(f"vasp_learn_add_{i + 1}_xyz_file"), xyz, format="extxyz")

    def check_env(self):


        if self.generation!=1 and self.config.get("restart") :
            utils.print("No need for initialization check.")
            utils.print_msg("--" * 4,
                            f"Restarting to train the potential function for the {self.generation}th generation.",
                            "--" * 4)

            return

        if self.config["current_job"]=="vasp":

            self.generation=0
            utils.copy(self.config["init_train_xyz"], self.vasp_learn_add_xyz_file)

            if self.config["vasp_job"] != 1:


                self.split_vasp_job_xyz(self.config["init_train_xyz"])
        elif self.config["current_job"]=="nep":
           

            utils.copy(self.config["init_train_xyz"], self.last_all_learn_calculated_xyz_file)
            # utils.copy(self.config["init_train_xyz"], self.last_all_learn_calculated_xyz_file )
            #如果势函数有效  直接先复制过来
        elif self.config["current_job"]=="gpumd":

            utils.copy(self.config["init_train_xyz"],self.nep_train_xyz_file )

            if os.path.exists(self.config["init_nep_txt"]):
                utils.copy(self.config["init_nep_txt"],
                            self.nep_nep_txt_file )
            else:
                raise FileNotFoundError("Starting task as gpumd requires specifying a valid potential function path!")
        else:
            raise ValueError("current_job can only be one of nep, gpumd, or vasp.")

    def read_config(self,config_path):
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"The file at {config_path} does not exist.")
        with open(config_path,"r",encoding="utf8") as f:


            self.config=YAML().load(f )

    def build_pred_params(self  ):
        nep=self.config["nep"]
        params=[]
        params.append("NepTrain")
        params.append("nep")

        params.append("--directory")
        params.append(self.pred_path)

        params.append("--in")
        params.append(os.path.abspath(nep.get("nep_in_path")))

        params.append("--train")
        params.append(self.all_learn_calculated_xyz_file)

        params.append("--nep")
        params.append(self.nep_nep_txt_file)

        params.append("--prediction")

        return " ".join(params)


    def build_nep_params(self  ):
        nep=self.config["nep"]
        params=[]
        params.append("NepTrain")
        params.append("nep")

        params.append("--directory")
        params.append(self.nep_path)

        params.append("--in")
        params.append(os.path.abspath(nep.get("nep_in_path")))

        params.append("--train")
        params.append(self.last_improved_train_xyz_file)

        params.append("--test")
        params.append(os.path.abspath(nep.get("test_xyz_path")))

        if self.config["nep"]["nep_restart"] and self.generation not in [1,len(self.config["gpumd"]["step_times"])+1]:
            #开启续跑
            #如果上一级的势函数路径有效  就传入一下续跑的参数

            if os.path.exists(self.last_nep_nep_restart_file):
                utils.print_tip("Start the restart mode!")


                params.append("--restart_file")
                params.append(self.last_nep_nep_restart_file)
                params.append("--continue_step")
                params.append(str(self.config["nep"]["nep_restart_step"]))

        return " ".join(params)
    def build_gpumd_params(self,n_job=1):
        gpumd=self.config["gpumd"]
        params=[]
        params.append("NepTrain")
        params.append("gpumd")

        params.append(os.path.abspath(gpumd.get("model_path")))

        params.append("--directory")

        params.append(self.gpumd_path)

        params.append("--in")
        params.append(os.path.abspath(gpumd.get("run_in_path")))
        params.append("--nep")
        params.append( self.nep_nep_txt_file)
        params.append("--time")
        params.append(str(gpumd.get("step_times")[self.generation-1]))

        params.append("--temperature")
        params.append( str(gpumd["temperature_every_step"][n_job] ))

        # params.append(" ".join([str(i) for i in gpumd["temperature_every_step"]]))


        params.append("--out")
        params.append(self.__getattr__(f"select_md_{n_job}_xyz_file"))

        if gpumd.get("filter_by_bonds",True):
            params.append("--filter")


        return " ".join(params)
    def build_select_params(self):
        select=self.config["select"]
        params=[]
        params.append("NepTrain")
        params.append("select")
        #总的
        # params.append(self.select_all_md_dummp_xyz_file)
        #分开
        params.append(self.__getattr__(f"select_md_*_xyz_file"))
        params.append("--nep")
        params.append( self.nep_nep_txt_file)

        params.append("--base")
        params.append( self.nep_train_xyz_file )
        params.append("--max_selected")
        params.append(str(select["max_selected"]))
        params.append("--min_distance")
        params.append(str(select["min_distance"]))
        params.append("--out")
        params.append(self.select_selected_xyz_file)




        return " ".join(params)
    def build_vasp_params(self,n_job=1):
        vasp=self.config["vasp"]
        params=[]
        params.append("NepTrain")
        params.append("vasp")

        if self.config["vasp_job"] == 1:

            if not os.path.exists(self.vasp_learn_add_xyz_file):
                return None
            params.append(self.vasp_learn_add_xyz_file)
        else:
            path=self.__getattr__(f"vasp_learn_add_{n_job}_xyz_file")
            if not os.path.exists(path):
                return None
            params.append(path)

        params.append("--directory")

        params.append(self.__getattr__(f"vasp_cache{n_job}_path"))


        params.append("-np")
        params.append(str(vasp["cpu_core"]))
        if vasp["kpoints_use_gamma"]:
            params.append("--gamma")

        if vasp["incar_path"]:

            params.append("--incar")
            params.append(os.path.abspath(vasp["incar_path"]))
        if vasp["use_k_stype"]=="kpoints":
            if vasp.get("kpoints"):
                params.append("-ka")
                if isinstance(vasp["kpoints"],list):
                    params.append(",".join([str(i) for i in vasp["kpoints"]]))
                else:
                    params.append(str(vasp["kpoints"]))
        else:

            if vasp.get("kspacing") :
                params.append("--kspacing")
                params.append(str(vasp["kspacing"]))
        params.append("--out")
        params.append( self.__getattr__(f"vasp_learn_calculated_{n_job}_xyz_file"))


        return " ".join(params)

    def select(self):
        utils.cat(self.__getattr__(f"select_md_*_xyz_file"),
                  self.select_all_md_dummp_xyz_file
                  )

        worker = LocalWorker()
        cmd=self.build_select_params()
        worker.sub_job(cmd, self.select_path)



        utils.cat(self.select_selected_xyz_file,
                  self.vasp_learn_add_xyz_file
                  )





    def sub_vasp(self):
        utils.print_msg("Beginning the execution of VASP for single-point energy calculations.")
        # break

        if not utils.is_file_empty(self.vasp_learn_add_xyz_file):

            if self.config["vasp_job"] != 1:
                # 这里分割下xyz 方便后面直接vasp计算
                self.split_vasp_job_xyz(self.vasp_learn_add_xyz_file)

            for i in range(self.config["vasp_job"]):
                cmd = self.build_vasp_params(i + 1)
                if cmd is None:
                    continue
                self.worker.sub_job(cmd, self.vasp_path, job_type="vasp")

            self.worker.wait()

            utils.cat(self.__getattr__(f"vasp_learn_calculated_*_xyz_file"),
                      self.all_learn_calculated_xyz_file

                      )
            if self.config.get("limit",{}).get("force") and not utils.is_file_empty(self.all_learn_calculated_xyz_file):
                bad_structure = []
                good_structure = []
                structures=ase_read(self.all_learn_calculated_xyz_file,":")
                for structure in structures:

                    if structure.calc.results["forces"].max() <= self.config.get("limit",{}).get("force"):
                        good_structure.append(structure)
                    else:
                        bad_structure.append(structure)

                ase_write(self.all_learn_calculated_xyz_file,good_structure,append=False,format="extxyz")
                if bad_structure:
                    ase_write(self.remove_by_force_xyz_file, bad_structure, append=False, format="extxyz")

        else:
            utils.print_warning("Detected that the calculation input file is empty, proceeding directly to the next step!")

            utils.cat(self.vasp_learn_add_xyz_file,
                      self.all_learn_calculated_xyz_file


                      )

    def sub_nep(self):
        utils.print_msg("--" * 4, f"Starting to train the potential function for the {self.generation}th generation.", "--" * 4)

        if not utils.is_file_empty(self.last_all_learn_calculated_xyz_file):


            if os.path.exists(self.last_nep_train_xyz_file):
                utils.cat([self.last_nep_train_xyz_file,
                           self.last_all_learn_calculated_xyz_file
                           ],
                          self.last_improved_train_xyz_file

                          )
            else:
                utils.copy(self.last_all_learn_calculated_xyz_file,
                            self.last_improved_train_xyz_file)
            utils.print_msg(f"Starting to train the potential function.")
            cmd = self.build_nep_params()
            self.worker.sub_job(cmd, self.nep_path, job_type="nep")
            self.worker.wait()
        else:
            utils.print_warning("The dataset has not changed, directly copying the potential function from the last time!")

            utils.copy_files(self.last_nep_path, self.nep_path)

    def sub_nep_pred(self):

        if utils.is_file_empty(self.nep_nep_txt_file):
            utils.print_msg(f"No potential function available, skipping prediction.")
            return
        if not utils.is_file_empty(self.all_learn_calculated_xyz_file):
            utils.print_msg(f"Starting to predict new dataset.")
            cmd = self.build_pred_params()
            self.worker.sub_job(cmd, self.pred_path, job_type="nep")
            self.worker.wait()
        else:
            utils.print_msg(f"The dataset has not changed, skipping prediction.")


    def sub_gpumd(self):
        utils.print_msg(f"Starting active learning.")
        for i in range(len(self.config["gpumd"]["temperature_every_step"])):
            cmd = self.build_gpumd_params(i)

            self.worker.sub_job(cmd, self.gpumd_path, job_type="gpumd")
        self.worker.wait()




    def start(self,config_path):
        utils.print_msg("Welcome to NepTrain automatic training!")

        self.read_config(config_path)
        self.check_env()

        self.worker=self.get_worker()

        self.manager.set_next(self.config.get("current_job"))

        while True:

            #开始循环
            job=next(self.manager)
            self.config["current_job"]=job
            self.save_restart()
            if job=="vasp":

                self.sub_vasp()

            elif job=="pred":

                self.sub_nep_pred()
                self.generation += 1

            elif job=="nep":

                self.sub_nep()
                if self.generation>len(self.config["gpumd"]["step_times"]):
                   utils.print_success("Training completed!")
                   break
            elif job=="select":
                self.select()

            else:
                self.sub_gpumd()


    def save_restart(self):
        with open("./restart.yaml","w",encoding="utf-8") as f:
            self.config["restart"]=True

            YAML().dump(self.config,f)

def train_nep(argparse):
    """
    首先检查下当前的进度 看从哪开始
    :return:
    """
    check_env()

    worker = NepTrainWorker()

    worker.start(argparse.config_path)
if __name__ == '__main__':
    train =NepTrainWorker()
    train.generation=1
    train.config["work_path"]="./cache"
    print(train.nep_path)

    print(train.__getattr__(f"vasp_learn_calculated_*_xyz_file"))



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import contextlib
import os

import numpy as np
from ase import Atoms

from NepTrain.nep_cpu import CpuNep

class Nep3Calculator:

    def __init__(self, model_file="nep.txt"):
        if not isinstance(model_file, str):
            model_file=str(model_file,encoding="utf-8")
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
                self.nep3 = CpuNep(model_file)
        self.element_list=self.nep3.get_element_list()
        self.type_dict = {e: i for i, e in enumerate(self.element_list)}




    def get_descriptors(self,structure):
        symbols = structure.get_chemical_symbols()
        _type = [self.type_dict[k] for k in symbols]
        _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
        _position = structure.get_positions().transpose(1, 0).reshape(-1).tolist()
        descriptor = self.nep3.get_descriptor(_type, _box, _position)
        descriptors_per_atom = np.array(descriptor).reshape(-1, len(structure)).T

        return descriptors_per_atom
    def get_structure_descriptors(self, structure):
        descriptors_per_atom=self.get_descriptors(structure)
        return descriptors_per_atom.mean(axis=0)

    def get_structures_descriptors(self,structures:[Atoms]):
        _types=[]
        _boxs=[]
        _positions=[]
        for structure in structures:
            symbols = structure.get_chemical_symbols()
            _type = [self.type_dict[k] for k in symbols]
            _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            _position = structure.get_positions().transpose(1, 0).reshape(-1).tolist()
            _types.append(_type)
            _boxs.append(_box)
            _positions.append(_position)
        descriptor = self.nep3.get_descriptors(_types, _boxs, _positions)

        return np.array(descriptor)




class DescriptorCalculator:
    def __init__(self, calculator_type="nep",**calculator_kwargs):
        self.calculator_type=calculator_type
        if calculator_type == "nep":
            self.calculator=Nep3Calculator(**calculator_kwargs)
        elif calculator_type == "soap":
            from dscribe.descriptors import SOAP

            self.calculator = SOAP(
                **calculator_kwargs,dtype="float32"
            )
        else:
            raise ValueError("calculator_type must be nep or soap")


    def get_structures_descriptors(self,structures:[Atoms]):

        if len(structures)==0:
            return np.array([])

        if self.calculator_type == "nep":
            return self.calculator.get_structures_descriptors(structures)
        else:

            return  np.array([self.calculator.create_single(structure).mean(0) for structure in structures])


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 15:53
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import subprocess

import numpy as np
from ase.calculators.vasp import Vasp
from ase.io import read as ase_read
from ase.io import write as ase_write
from NepTrain import Config
def write_to_xyz(vaspxml_path, save_path, Config_type, append=True):

    atoms_list = []
    atoms = ase_read(vaspxml_path, index=":")
    index = 1
    for atom in atoms:
        xx, yy, zz, yz, xz, xy = -atom.calc.results['stress'] * atom.get_volume()  # *160.21766
        atom.info['virial'] = np.array([(xx, xy, xz), (xy, yy, yz), (xz, yz, zz)])

        atom.calc.results['energy'] = atom.calc.results['free_energy']

        atom.info['Config_type'] = Config_type + str(index)
        atom.info['Weight'] = 1.0
        del atom.calc.results['stress']
        del atom.calc.results['free_energy']
        atoms_list.append(atom)
        index += 1

    ase_write(save_path, atoms_list, format='extxyz', append=append)
    return atoms_list
class VaspInput(Vasp):


    def __init__(self,*args,**kwargs):

        super(VaspInput,self).__init__(*args,**kwargs)
        self.input_params["setups"] = {"base": "recommended"}
        self.input_params["pp"] = ''

        os.environ[self.VASP_PP_PATH] = os.path.expanduser(Config.get("environ", "potcar_path"))


        # self.converged
    def _run(self, command=None, out=None, directory=None):
        """Method to explicitly execute VASP"""
        if command is None:
            command = self.command
        if directory is None:
            directory = self.directory

        errorcode = subprocess.call(command,
                                    shell=True,
                                    stdout=out,
                                    stderr=out,
                                    cwd=directory)

        return errorcode
if __name__ == '__main__':
    vasp=VaspInput()




    atoms=ase_read("./POSCAR",format='vasp')
    vasp.read_incar("./INCAR")
    vasp.calculate(atoms,('energy'))
    print(vasp.results)
    print(vasp.atoms.info)
    print(atoms.calc.results)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/13 19:36
# @Author  : 兵
# @email    : 1747193328@qq.com
import os.path
from pathlib import Path

import numpy as np


# from joblib import Parallel, delayed

from NepTrain import utils
from ase.io import read as ase_read
from ase.io import write as ase_write
from .select import select_structures, filter_by_bonds, farthest_point_sampling
from ..gpumd.plot import plot_md_selected

from ..nep.calculator import DescriptorCalculator


def run_select(argparse):
    import matplotlib.pyplot as plt
    map_path_index=[]
    all_trajectory=[]
    plot_config=[]
    trajectory_structures=[]
    for index,_path in enumerate(argparse.trajectory_paths):

        if utils.is_file_empty(_path):
            utils.print_warning(f"An invalid file path was provided: {argparse.trajectory_path}.")
            continue

        utils.print_msg(f"Reading trajectory {_path}")

        trajectory=ase_read(_path,":",format="extxyz")
        map_path_index.append(np.full(len(trajectory),index))
        trajectory_structures.extend(trajectory)
    map_path_index=np.concatenate(map_path_index)


    if utils.is_file_empty(argparse.base):
        base_train=[]
    else:
        base_train=ase_read(argparse.base,":",format="extxyz")

    if utils.is_file_empty(argparse.nep):
        utils.print_msg("An invalid path for nep.txt was provided, using SOAP descriptors instead.")
        species=set()
        for atoms in trajectory+base_train:
            for i in atoms.get_chemical_symbols():
                species.add(i)
        kwargs_dict={
            "species":list(species),
            "r_cut":argparse.r_cut,
            "n_max": argparse.n_max,
            "l_max": argparse.l_max

        }

        descriptor =DescriptorCalculator("soap",**kwargs_dict)

    else:
        descriptor =DescriptorCalculator("nep", model_file=argparse.nep)

    utils.print_msg("Start generating structure descriptor, please wait")
    train_structure_des =descriptor.get_structures_descriptors(base_train)
    trajectory_structure_des=[]
    uniqe_index = np.unique(map_path_index)

    # 使用 'viridis' colormap 从色图中获取颜色
    cmap = plt.cm.viridis  # 你也可以选择其他色图，例如 'plasma', 'inferno', 'cividis' 等
    num_colors = len(uniqe_index)

    # 创建颜色数组
    colors = [cmap(i / num_colors) for i in range(num_colors)]
    for index  in uniqe_index:
        indices = np.where(map_path_index == index)[0]
        trajectory_structure=[trajectory_structures[i] for i in indices]
        new =descriptor.get_structures_descriptors(trajectory_structure)
        trajectory_structure_des.append(new)
        label = Path(argparse.trajectory_paths[index]).name

        plot_config.append((new,label, colors[index]))

    utils.print_msg("Starting to select points, please wait...")
    new_structure_des=np.vstack(trajectory_structure_des)


    selected_i =farthest_point_sampling(new_structure_des,argparse.max_selected,argparse.min_distance,selected_data=train_structure_des)

    selected_structures=[]

    base_dir=os.path.dirname(argparse.out_file_path)

    utils.remove_file_by_re(os.path.join(base_dir,"selected*.xyz"))

    for i in selected_i:
        file_index=map_path_index[i]

        fila_name = Path(argparse.trajectory_paths[file_index]).name
        save_path=os.path.join(base_dir,f"selected_{fila_name}.xyz")


        structure = trajectory_structures[i]
        ase_write(save_path,structure,append=True)
        selected_structures.append(structure)
    selected_des=new_structure_des[selected_i,:]


    utils.print_msg(f"Obtained {len(selected_i)} structures." )
    ase_write(argparse.out_file_path, selected_structures)
    png_path=os.path.join(os.path.dirname(argparse.out_file_path),"selected.png")
    plot_md_selected(train_structure_des,
                     plot_config,
                     selected_des,
                       png_path ,
                     argparse.decomposition
                     )
    utils.print_msg(f"The point selection distribution chart is saved to {png_path}." )
    utils.print_msg(f"The selected structures are saved to {argparse.out_file_path}." )




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/13 19:09
# @Author  : 兵
# @email    : 1747193328@qq.com



from .select import select_structures,process_trajectory,filter_by_bonds

from .run import run_select




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 16:01
# @Author  : 兵
# @email    : 1747193328@qq.com
import glob
import os
import shutil
import subprocess
import traceback
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Generator, Union

from ase.io import read as ase_read
from rich import get_console
from rich.progress import track

#前面几个0是为了让元素编号和索引对的上 避免了见一
radius_table = {'H': 0.31, 'He': 0.28, 'Li': 1.28, 'Be': 0.96,
                'B': 0.85, 'C': 0.76, 'N': 0.71, 'O': 0.66,
                'F': 0.57, 'Ne': 0.58, 'Na': 1.66, 'Mg': 1.41,
                'Al': 1.21, 'Si': 1.11, 'P': 1.07, 'S': 1.05,
                'Cl': 1.02, 'Ar': 1.06, 'K': 2.03, 'Ca': 1.76,
                'Sc': 1.7, 'Ti': 1.6, 'V': 1.53, 'Cr': 1.39,
                'Mn': 1.39, 'Fe': 1.32, 'Co': 1.26, 'Ni': 1.24,
                'Cu': 1.32, 'Zn': 1.22, 'Ga': 1.22, 'Ge': 1.2,
                'As': 1.19, 'Se': 1.2, 'Br': 1.2, 'Kr': 1.16,
                'Rb': 2.2, 'Sr': 1.95, 'Y': 1.9, 'Zr': 1.75,
                'Nb': 1.64, 'Mo': 1.54, 'Tc': 1.47, 'Ru': 1.46,
                'Rh': 1.42, 'Pd': 1.39, 'Ag': 1.45, 'Cd': 1.44,
                'In': 1.42, 'Sn': 1.39, 'Sb': 1.39, 'Te': 1.38,
                'I': 1.39, 'Xe': 1.4, 'Cs': 2.44, 'Ba': 2.15,
                'La': 2.07, 'Ce': 2.04, 'Pr': 2.03, 'Nd': 2.01,
                'Pm': 1.99, 'Sm': 1.98, 'Eu': 1.98, 'Gd': 1.96,
                'Tb': 1.94, 'Dy': 1.92, 'Ho': 1.92, 'Er': 1.89,
                'Tm': 1.9, 'Yb': 1.87, 'Lu': 1.87, 'Hf': 1.75,
                'Ta': 1.7, 'W': 1.62, 'Re': 1.51, 'Os': 1.44,
                'Ir': 1.41, 'Pt': 1.36, 'Au': 1.36, 'Hg': 1.32,
                'Tl': 1.45, 'Pb': 1.46, 'Bi': 1.48, 'Po': 1.4,
                'At': 1.5, 'Rn': 1.5, 'Fr': 2.6, 'Ra': 2.21,
                'Ac': 2.15, 'Th': 2.06, 'Pa': 2.0, 'U': 1.96,
                'Np': 1.9, 'Pu': 1.87, 'Am': 1.8, 'Cm': 1.69,
                'Bk': 1.5, 'Cf': 1.5, 'Es': 1.5, 'Fm': 1.5,
                'Md': 1.5, 'No': 1.5, 'Lr': 1.5, 'Rf': 1.5,
                'Db': 1.5, 'Sg': 1.5, 'Bh': 1.5, 'Hs': 1.5,
                'Mt': 1.5, 'Ds': 1.5, 'Rg': 1.5, 'Cn': 1.5,
                'Nh': 1.5, 'Fl': 1.5, 'Mc': 1.5, 'Lv': 1.5,
                'Ts': 1.5, 'Og': 1.5}

def print(*msg, **kwargs):

    get_console().print(f"[{datetime.now()}] -- ",*msg, **kwargs)


def print_warning(*msg):
    print(*msg, style="#fc5531")

def print_msg(*msg):
    print(*msg )

def print_tip(*msg):
    print(*msg)

def print_success(*msg):
    print(*msg, style="green")


def merge_yaml(yaml_a, yaml_b):
    result = yaml_a.copy()  # 复制a的内容
    for key, value in yaml_b.items():  # 遍历b的键值对
        if key in yaml_a and isinstance(yaml_a[key], dict) and isinstance(yaml_b[key], dict):
            result[key] = merge_yaml(yaml_a[key], yaml_b[key])  # 递归合并字典
        else:
            result[key] = value  # 否则直接覆盖
    return result


def get_config_path():
    return os.path.join(os.path.expanduser('~'),".NepTrain")

def verify_path(path):
    if not os.path.exists(os.path.expanduser(path)):

        os.makedirs(os.path.expanduser(path))

def copy(rc, dst,   follow_symlinks=True):

    parent_path=(os.path.dirname(dst))
    if not os.path.exists(parent_path):
        os.makedirs(parent_path)
    shutil.copy(rc, dst,  follow_symlinks=follow_symlinks)
def copy_files(src_dir, dst_dir):
    # 遍历源目录中的所有文件
    for filename in os.listdir(src_dir):
        src_file = os.path.join(src_dir, filename)
        dst_file = os.path.join(dst_dir, filename)

        # 确保是文件而不是目录
        if os.path.isfile(src_file):
            # 复制文件
            shutil.copy2(src_file, dst_file)

def remove_file_by_re(src):


    src_file_list =  glob.glob(src)
    for file in src_file_list:
        Path(file).unlink()



def cat(files,out_file):


    # 获取所有以 'file' 开头的文件名
    if isinstance(files,str):
        file_list = glob.glob(files)
    else:
        file_list = files

    # 打开目标文件用于写入
    with open(out_file, 'wb') as outfile:
        for filename in file_list:

            with open(filename, 'rb') as infile:
                # 读取文件内容并写入到目标文件
                outfile.write(infile.read())


@contextmanager
def cd(path: Union[str, Path]) -> Generator:
    """


        with cd("/my/path/"):
            do_something()

    Args:
        path: Path to cd to.
    """
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)
def iter_path_to_atoms(glob_strs: list,show_progress=True,**kkwargs):
    def decorator(func):
        def wrapper(path: Path | str, *args, **kwargs):
            if isinstance(path, str):
                path = Path(path)
            paths=[]
            if path.is_dir():
                parent = path
                for glob_str in glob_strs:
                    for i in parent.glob(glob_str):
                        paths.append(i)
            else:
                paths = [path]
            result =[]

            filter_path_list=[]

            for i in paths:
                try:
                    atoms=ase_read(i.as_posix(),index=":")
                except Exception as e:
                    print_warning(f"文件：{i.as_posix()}读取错误!报错原因：{e}")
                    continue
                if isinstance(atoms,list):

                    filter_path_list.extend(atoms)
                else:
                    filter_path_list.append(atoms)
            if len(filter_path_list)==0:
                print_warning(f"Structure file not found: {path}")
                return result
            if show_progress:
                iter_obj=track(filter_path_list,
                              **kkwargs
                              )
            else:
                iter_obj=filter_path_list

            for i in iter_obj:

                try:
                    result.append(func(i, *args, **kwargs))
                except KeyboardInterrupt:
                    return result
                except Exception as e:
                    print(traceback.format_exc())
                    print_warning(traceback.format_exc())
                    print_warning(e)
                    pass
            return result
        return wrapper

    return decorator




def get_command_result(cmd):
    try:
        # 使用 subprocess 调用 which 命令，并捕获输出

        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        # 检查命令是否成功执行

        if result.returncode == 0:
            # 返回命令的路径
            return result.stdout.strip()
        else:
            # 如果命令未找到，返回 None 或抛出异常
            return None
    except Exception as e:

        return None

def is_file_empty(file_path):
    # 检查文件是否存在

    if file_path is None:
        return True

    if not os.path.exists(file_path):
        print_warning(f"The file {file_path} does not exist.")
        return True
        # raise FileNotFoundError(f"文件 {file_path} 不存在。")

    # 检查文件大小
    return os.path.getsize(file_path) == 0
def is_diff_path(path,path1):
    return os.path.abspath(path)!=os.path.abspath(path1)

def split_list(lst, n):
    k, m = divmod(len(lst), n)
    return [lst[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n)]





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
"""Pymatgen package configuration."""

from __future__ import annotations

import platform
import sys
from distutils.ccompiler import get_default_compiler
import os
import subprocess
import pybind11
from pybind11.setup_helpers import Pybind11Extension
from setuptools import  Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

# 获取 pybind11 的 include 路径
pybind11_include = pybind11.get_include()

# 检测当前编译器
compiler = get_default_compiler()

# 设定编译选项
extra_compile_args = []
extra_link_args = []


def check_openmp_support():
    code = """
    #include <omp.h>
    #include <stdio.h>
    int main() {
        #ifdef _OPENMP
        return 0;
        #else
        return 1;
        #endif
    }
    """
    with open("test_openmp.c", "w") as f:
        f.write(code)

    compiler = os.environ.get("CC", "gcc")  # 默认使用 gcc，用户可通过 CC 环境变量指定编译器
    try:
        subprocess.run([compiler, "-fopenmp", "test_openmp.c", "-o", "test_openmp"],
                       check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result = subprocess.run(["./test_openmp"], check=True)
        return result.returncode == 0
    except subprocess.CalledProcessError:
        return False
    finally:
        os.remove("test_openmp.c")
        for o in ["test_openmp","test_openmp.exe","test_openmp.bin","test_openmp.o"]:
            if os.path.exists(o):
                os.remove(o)

# 检测结果
openmp_supported = check_openmp_support()




if platform.system() == 'Windows' and compiler == 'msvc':  # 对于 MSVC 编译器（Windows）
    extra_compile_args = ['/O2', '/std:c++17' ]
    if openmp_supported:
        extra_compile_args.append('/openmp')
        extra_link_args.append('/openmp')

elif platform.system() != 'Windows' and compiler != 'msvc':  # 对于 GCC 或 Clang 编译器（Linux/macOS）
    extra_compile_args = ['-O3', '-std=c++17' ]
    if openmp_supported:
        extra_compile_args.append('-fopenmp')
        extra_link_args.append('-fopenmp')
# 定义扩展模块
ext_modules = [
    Pybind11Extension(
        "NepTrain.nep_cpu",  # 模块名
        ["src/nep_cpu/nep_cpu.cpp"],  # 源文件
        include_dirs=[
            pybind11_include,
            # "src/nep_cpu"
        ],
        extra_compile_args=extra_compile_args,  # 编译选项
        extra_link_args=extra_link_args,
        language="c++",  # 指定语言为 C++
    ),
]

# 自定义 build_ext 命令，确保兼容性
class BuildExt(build_ext):
    def build_extensions(self):
        # 设置编译器标准为 C++17
        ct = self.compiler.compiler_type
        opts = [ ]
        for ext in self.extensions:
            ext.extra_compile_args = opts + ext.extra_compile_args
        build_ext.build_extensions(self)



setup(
    author="Chen Cheng bing",
cmdclass={'build_ext': BuildExt},
    # include_dirs=[np.get_include()],
packages=find_packages(),
ext_modules=ext_modules,
zip_safe=False,
include_package_data=True,  # 确保包含额外的文件（如 .so）
)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 19:56
# @Author  : 兵
# @email    : 1747193328@qq.com
import os.path


import numpy as np
from ase.io import read as ase_read




from NepTrain import utils

from .utils import  read_thermo





def plot_md_selected(train_des,md_des,selected_des, save_path,decomposition="pca"):
    # 画一下图
    from matplotlib import pyplot as plt

    config = [
        # (文件名,图例,图例颜色)

    ]
    if train_des is not None and train_des.size!=0:
        config.append((train_des, "base dataset","gray"))

    if md_des is not None:
        if isinstance(md_des,np.ndarray) and md_des.size!=0:
            config.append((md_des, 'new dataset', "#07cd66"))
        elif isinstance(md_des,list) :
            config.extend(md_des)
    if selected_des is not None  and selected_des.size!=0:
        config.append((selected_des,'selected', "red"))

    fit_data = []

    for info in config:

        # atoms_list_des = np.vstack([get_descriptors(i, nep_txt_path) for i in atoms_list])

        atoms_list_des= info[0]
        # print(atoms_list_des.shape)


        fit_data.append(atoms_list_des)
    if decomposition=="pca":
        from sklearn.decomposition import PCA

        reducer = PCA(n_components=2)
    else:
        from umap import UMAP

        reducer = UMAP(n_components=2)

    reducer.fit(np.vstack(fit_data))
    fig = plt.figure()
    for index, array in enumerate(fit_data):
        proj = reducer.transform(array)
        plt.scatter(proj[:, 0], proj[:, 1], label=config[index][1], c=config[index][2])
    leg_cols=len(config)//3 or 1
    plt.legend(ncols=leg_cols)
    # plt.axis('off')
    plt.savefig(save_path)
    plt.close(fig)









def plot_energy(thermo_path,natoms=1):
    from matplotlib import pyplot as plt

    potential_energy = read_thermo(thermo_path, natoms)



    fig = plt.figure()
    plt.plot(list(range(len(potential_energy ))), potential_energy)



    plt.savefig(os.path.join(os.path.dirname(thermo_path),"md_energy.png"), dpi=300)
    plt.close(fig)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/26 19:39
# @Author  : 兵
# @email    : 1747193328@qq.com
import numpy as np


def read_thermo(filename: str, natoms: int = 1) -> np.ndarray:
    """读取能量用于画图
    """
    data = np.loadtxt(filename)

    energy = data[:,2]

    assert natoms > 0, 'natoms must be positive'

    energy /= natoms
    return energy


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 20:51
# @Author  : 兵
# @email    : 1747193328@qq.com
import re
import subprocess
import sys
import time
from threading import Thread

from NepTrain import utils


class Worker():
    pass
    def __init__(self, mode):
        self.mode = mode

        self._running=False

    @property
    def running(self):
        return self._running

    def sub_job(self,command,job_path):
        raise NotImplementedError

    def wait(self):
        pass

class LocalWorker(Worker):
    def __init__(self,  ):
        super().__init__("local")
    def sub_job(self,command,job_path,**kwargs):
        utils.verify_path(job_path)

        # with  open("job.out", "w") as f_std, open("job.err", "w", buffering=1) as f_err:
        errorcode = subprocess.call(command,
                                    shell=True,
                                    stdout=sys.stdout,
                                    stderr=sys.stderr,
                                    cwd=job_path)



class SlurmWorker(Worker):
    def __init__(self,vasp_sh,gpumd_sh  ):
        super().__init__("slurm")
        self.vasp_sh = vasp_sh
        self.gpumd_sh = gpumd_sh
        #创建一个线程  定时检查任务状态？
        self.job_id=[]
        self._thread=Thread(target=self.check_job_state)
    def sub_job(self,command,job_path,job_type="vasp"):
        utils.verify_path(job_path)

        if job_type == "vasp":
            job_command=["sbatch",self.vasp_sh,command]
        else:
            job_command=["sbatch",self.gpumd_sh,command]
        # print(command)

        result = subprocess.run(job_command, capture_output=True, text=True, check=True,cwd=job_path)
        job_id=int(result.stdout.replace("Submitted batch job ",""))
        self.job_id.append(job_id)
        utils.print_msg(f"Task submitted: {job_id}",)


    def wait(self):
        utils.print_msg("Waiting for all tasks to finish...")
        while  self.job_id:
            for job in self.job_id.copy():

                if not self.check_job_state(job):
                    self.job_id.remove(job)
            time.sleep(5)
        utils.print_success("All tasks have finished computation.")

    def check_job_state(self,job_id):

        result = subprocess.run(['squeue','-j',f"{job_id}"], capture_output=True, text=True, check=True)
        match = re.search(r'JOBID.*?(\d+) ', result.stdout, re.S)
        #
        # 如果找到匹配项，打印作业ID
        if match:
            # job_id = match.group(1)
            return True


        else:
            return False

